from typing import Self

import attrs
import numpy as np

from spaghettree import safe
from spaghettree.logger import logger


@attrs.define
class AdjMat:
    mat: np.ndarray = attrs.field()
    node_map: dict[int, str] = attrs.field()
    communities: list[int] = attrs.field()
    comm_map: dict[int, str] = attrs.field(factory=dict)

    @classmethod
    @safe
    def from_call_tree(cls, call_tree: dict[str, list[str]]) -> Self:
        logger.debug(f"{call_tree = }")
        ent_idx: dict[str, int] = {node: i for i, node in enumerate(call_tree)}
        node_map: dict[int, str] = {idx: ent_name for ent_name, idx in ent_idx.items()}

        adj_mat = AdjMat._create_adj_map(call_tree, ent_idx)

        return cls(adj_mat, node_map, list(node_map.keys()))

    @classmethod
    @safe
    def from_call_tree_no_optimisation(cls, call_tree: dict[str, list[str]]) -> Self:
        logger.debug(f"{call_tree = }")
        ent_idx: dict[str, int] = {node: i for i, node in enumerate(call_tree)}
        node_map: dict[int, str] = {idx: ent_name for ent_name, idx in ent_idx.items()}

        modules = [".".join(k.split(".")[:-1]) for k in call_tree]
        unique_mods = list(dict.fromkeys(modules))
        mod_map = {name: idx for idx, name in enumerate(unique_mods)}
        communities = [mod_map[name] for name in modules]

        adj_mat = AdjMat._create_adj_map(call_tree, ent_idx)

        return cls(adj_mat, node_map, communities, comm_map={v: k for k, v in mod_map.items()})

    @staticmethod
    def _create_adj_map(call_tree: dict[str, list[str]], ent_idx: dict[str, int]) -> np.ndarray:
        n = len(ent_idx)
        adj_mat = np.zeros((n, n), dtype=int)

        for caller, called in call_tree.items():
            for call in called:
                src_idx = ent_idx[caller]
                dst_idx = ent_idx[call]
                adj_mat[src_idx, dst_idx] += 1
        return adj_mat



from __future__ import annotations

import os
import pathlib
from copy import deepcopy

import libcst as cst
import numpy as np
from tqdm import tqdm

from spaghettree import safe
from spaghettree.domain.adj_mat import AdjMat
from spaghettree.domain.entities import EntityCST
from spaghettree.domain.visitors import EntityLocation, OnePassVisitor
from spaghettree.logger import logger


def str_to_cst(code: str) -> cst.Module:
    return cst.parse_module(code)


def cst_to_str(node: cst.CSTNode) -> str:
    return cst.Module([]).code_for_node(node)


@safe
def extract_entities_and_locations(
    src_code: dict[str, str],
) -> tuple[dict[str, EntityCST], dict[str, EntityLocation]]:
    def find_common_prefix(paths: list[str]) -> str:
        logger.debug(f"{paths = }")
        return str(pathlib.Path(os.path.commonpath(paths)).parent)

    def get_module_name(path: str, root: str) -> str:
        return os.path.splitext(path.removeprefix(root))[0].replace("/", ".").strip(".")

    common_prefix = find_common_prefix(src_code.keys())
    logger.debug(f"{common_prefix = }")

    entities: dict[str, EntityCST] = {}
    locations: dict[str, EntityLocation] = {}

    for path, data in tqdm(src_code.items(), "creating objects"):
        tree = cst.metadata.MetadataWrapper(str_to_cst(data))
        module_name = get_module_name(path, common_prefix)
        visitor = OnePassVisitor(module_name)
        tree.visit(visitor)

        entities.update(visitor.entities)
        locations.update(visitor.locations)

        import_map = {
            i.as_name: f"{i.module}.{i.as_name}" if i.module != i.as_name else i.module
            for i in visitor.imports
        }
        ent_map = {ent.name.split(".")[-1]: ent.name for ent in entities.values()}

        logger.debug(f"{import_map = }")
        logger.debug(f"{ent_map = }")

        for ent in visitor.entities.values():
            ent.resolve_calls(import_map, ent_map).add_referenced_imports(visitor.imports)

        logger.debug(f"{visitor.entities = }")
        logger.debug(f"{visitor.imports = }")

    logger.debug(f"{entities = }")
    logger.debug(f"{locations = }")
    return entities, locations


@safe
def filter_non_native_calls(
    entities: dict[str, EntityCST],
) -> dict[str, EntityCST]:
    logger.debug(f"{entities = }")
    entities = deepcopy(entities)
    return {
        name: ent.filter_native_calls(entities).resolve_native_imports()
        for name, ent in entities.items()
    }


@safe
def create_call_tree(entities: dict[str, EntityCST]) -> dict[str, list[str]]:
    logger.debug(f"{entities = }")
    return {name: ent.get_call_tree_entries() for name, ent in entities.items()}


@safe
def pair_exclusive_calls(adj_mat: AdjMat) -> AdjMat:
    adj_mat = deepcopy(adj_mat)
    matrix: np.ndarray = adj_mat.mat
    communities: list[int] = adj_mat.communities.copy()

    # make it so we don't weight by call count yet
    adj_bin = (matrix > 0).astype(bool)
    communities: np.ndarray = np.array(communities, dtype=int)

    changed = True
    while changed:
        changed = False

        out_deg = adj_bin.sum(axis=1)
        in_deg = adj_bin.sum(axis=0)

        rows, cols = np.where((out_deg == 1)[:, None] & adj_bin & (in_deg == 1))

        for a, b in zip(rows, cols, strict=False):
            if communities[b] != communities[a]:
                communities[communities == communities[b]] = communities[a]
                changed = True

    adj_mat.communities = communities.tolist()
    return adj_mat



from __future__ import annotations

import functools
from collections.abc import Callable
from types import TracebackType
from typing import (
    Any,
    Literal,
    ParamSpec,
    Self,
    TypeVar,
)

import attrs

T = TypeVar("T")
P = ParamSpec("P")


@attrs.define
class Ok:
    inner: Any = attrs.field(default=None)

    def is_ok(self) -> Literal[True]:
        return True

    def and_then(self, func: Callable[[T], Result]) -> Result:
        return func(self.inner)

    def unwrap(self) -> T:
        return self.inner


@attrs.define
class Err:
    input_args: Any = attrs.field(repr=False)
    error: Exception | None = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(Exception)),
    )
    err_type: BaseException = attrs.field(init=False, repr=False)
    err_msg: str = attrs.field(init=False, repr=False)
    details: list[dict[str, Any]] = attrs.field(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        self.err_type = type(self.error)
        self.err_msg = str(self.error)
        if self.error:
            self.details = self._extract_details(self.error.__traceback__)

    def _extract_details(self, tb: TracebackType | None) -> list[dict[str, Any]]:
        trace_info = []
        while tb:
            frame = tb.tb_frame
            trace_info.append(
                {
                    "file": frame.f_code.co_filename,
                    "func": frame.f_code.co_name,
                    "line_no": tb.tb_lineno,
                    "locals": frame.f_locals,
                },
            )
            tb = tb.tb_next
        return trace_info

    def is_ok(self) -> Literal[False]:
        return False

    def and_then(self, _: Callable[[T], Result]) -> Self:
        return self

    def unwrap(self) -> None:
        raise self.error


Result = Ok | Err


def safe[**P, T](func: Callable[P, T]) -> Callable[P, Result]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result:
        try:
            return Ok(func(*args, **kwargs))
        except Exception as e:  # noqa: BLE001
            return Err((args, kwargs), e)

    return wrapper


__all__ = [
    "Err",
    "Ok",
    "Result",
]



from __future__ import annotations

import glob
import os
import subprocess
from pathlib import Path
from typing import Protocol, runtime_checkable

import attrs
import black
import isort
from ruff.__main__ import find_ruff_bin

from spaghettree import Err, Ok, Result, safe


@runtime_checkable
class IOProtocol(Protocol):
    @safe
    def list_files(self, root: str | Path, *, recursive: bool = True) -> list[str]: ...

    @safe
    def read(self, path: str) -> str: ...

    @safe
    def read_files(self, root: str | Path) -> Result: ...

    @safe
    def write(self, modified_code: str, filepath: str, *, format_code: bool = True) -> None: ...

    def write_files(self, src_code: dict[str, str], ruff_root: str | None = None) -> Result: ...


@attrs.define
class IOWrapper:
    @safe
    def list_files(self, root: str | Path, *, recursive: bool = True) -> list[str]:
        root = os.path.abspath(str(root))
        return sorted(glob.glob(f"{root}/**/**.py", recursive=recursive))

    @safe
    def read(self, path: str) -> str:
        with open(path) as f:
            return f.read()

    def read_files(self, root: str | Path) -> Result:
        paths_res = self.list_files(root)
        if not paths_res.is_ok():
            return paths_res
        paths = paths_res.inner

        results, fails = {}, {}
        for path in paths:
            res = self.read(path)
            if res.is_ok():
                results[path] = res.inner
            else:
                fails[path] = res

        if fails:
            return Err(fails)
        return Ok(results)

    @safe
    def write(self, modified_code: str, filepath: str, *, format_code: bool = True) -> None:
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        with open(filepath, "w") as f:
            f.write(format_code_str(modified_code))
        if format_code:
            self._run_ruff(filepath)

    def write_files(self, src_code: dict[str, str], ruff_root: str | None = None) -> Result:
        results, fails = {}, {}

        for filepath, modified_code in src_code.items():
            if ruff_root is not None:
                # format all at the end instead
                res = self.write(modified_code, filepath, format_code=False)
            else:
                res = self.write(modified_code, filepath, format_code=True)

            if res.is_ok():
                results[filepath] = res.inner
            else:
                fails[filepath] = res

        if ruff_root:
            self._run_ruff(ruff_root)
        if fails:
            return Err(fails)
        return Ok(results)

    def _run_ruff(self, path: str) -> None:
        subprocess.run([find_ruff_bin(), "check", "--fix", str(path)], check=True)  # noqa: S603
        subprocess.run([find_ruff_bin(), "format", str(path)], check=True)  # noqa: S603


@attrs.define
class FakeIOWrapper:
    files: dict = attrs.field(factory=dict)

    @safe
    def list_files(self, root: str | Path, *, recursive: bool = True) -> list[str]:
        if recursive:
            return sorted([f for f in self.files if root in f and f.endswith(".py")])
        return sorted(
            [
                f
                for f in self.files
                if f.removeprefix(root).lstrip("/").split("/")[0].endswith(".py")
            ]
        )

    @safe
    def read(self, path: str) -> str:
        return self.files[path]

    def read_files(self, root: str | Path) -> Result:
        paths_res = self.list_files(root)
        if not paths_res.is_ok():
            return paths_res
        paths = paths_res.inner

        results, fails = {}, {}
        for path in paths:
            res = self.read(path)
            if res.is_ok():
                results[path] = res.inner
            else:
                fails[path] = res

        if fails:
            return Err(fails)
        return Ok(results)

    @safe
    def write(self, modified_code: str, filepath: str, *, format_code: bool = True) -> None:
        self.files[filepath] = format_code_str(modified_code) if format_code else modified_code

    def write_files(self, src_code: dict[str, str], ruff_root: str | None = None) -> Result:
        results, fails = {}, {}

        for filepath, modified_code in src_code.items():
            if ruff_root is not None:
                res = self.write(modified_code, filepath)

            if res.is_ok():
                results[filepath] = res.inner
            else:
                fails[filepath] = res

        if fails:
            return Err(fails)
        return Ok(results)


def format_code_str(code: str) -> str:
    return black.format_str(isort.code(code), mode=black.FileMode())



from collections import defaultdict
from copy import deepcopy

import attrs
import numpy as np

from spaghettree import safe
from spaghettree.domain.adj_mat import AdjMat
from spaghettree.logger import logger


@safe
def optimise_communities(adj_mat: AdjMat) -> AdjMat:
    valid_merges = get_merge_pairs(adj_mat)
    logger.debug(f"{get_dwm(adj_mat.mat, adj_mat.communities) = }")

    while valid_merges:
        to_merge = remove_overlapping_pairs(valid_merges)
        adj_mat.communities = apply_merges(adj_mat.communities, to_merge)
        valid_merges = get_merge_pairs(adj_mat)

    logger.debug(f"{get_dwm(adj_mat.mat, adj_mat.communities) = }")
    logger.debug(f"{adj_mat.communities = }")
    return adj_mat


@safe
def merge_single_entity_communities_if_no_gain_penalty(adj_mat: AdjMat) -> AdjMat:
    communities = np.array(adj_mat.communities)
    base_score = get_dwm(adj_mat.mat, communities)

    grouped: defaultdict[int, list[tuple[int, str]]] = defaultdict(list)

    for idx, ent_name in adj_mat.node_map.items():
        grouped[communities[idx]].append((idx, ent_name))

    updated, min_for_dir = {}, {}

    for comm, items in grouped.items():
        if len(items) == 1:
            num, name = items[0]
            dirname = ".".join(name.split(".")[:-1])
            min_for_dir[dirname] = min(num, min_for_dir.get(dirname, num))
            updated[comm] = min_for_dir[dirname]

    merge_pairs = []

    for c2, c1 in updated.items():
        merged_communities = communities.copy()
        merged_communities[merged_communities == c2] = c1
        score = get_dwm(adj_mat.mat, merged_communities)

        if gain := (score - base_score) >= 0:
            merge_pairs.append(PossibleMerge(c1, c2, gain))

    adj_mat.communities = apply_merges(adj_mat.communities, merge_pairs)
    return adj_mat


@attrs.define(eq=True, frozen=True)
class PossibleMerge:
    c1: int = attrs.field()
    c2: int = attrs.field()
    gain: float = attrs.field()


def get_merge_pairs(adj_mat: AdjMat) -> list[PossibleMerge]:
    communities = np.array(adj_mat.communities)
    unique_comms = np.unique(communities)
    base_score = get_dwm(adj_mat.mat, communities)

    merge_scores = []

    for i, c1 in enumerate(unique_comms):
        for c2 in unique_comms[i + 1 :]:
            merged_communities = communities.copy()

            merged_communities[merged_communities == c2] = c1
            score = get_dwm(adj_mat.mat, merged_communities)
            gain = score - base_score
            if gain > 0:
                merge_scores.append(PossibleMerge(c1, c2, gain))

    logger.debug(f"{merge_scores = }")
    return merge_scores


def remove_overlapping_pairs(pairs: list[PossibleMerge]) -> list[PossibleMerge]:
    pairs = sorted(pairs, key=lambda x: x.gain, reverse=True)

    selected, seen = [], set()

    for pair in pairs:
        if pair.c1 not in seen and pair.c2 not in seen:
            selected.append(pair)
            seen.add(pair.c1)
            seen.add(pair.c2)

    logger.debug(f"{selected = }")
    return selected


def apply_merges(communities: list[int], pairs: list[PossibleMerge]) -> list[int]:
    communities = np.array(communities)

    for pair in pairs:
        communities[communities == pair.c2] = pair.c1
    return communities.tolist()


def get_dwm(mat: np.ndarray, communities: list[int]) -> float:
    out_degree = mat.sum(axis=0)
    in_degree = mat.sum(axis=1)
    total_edges = out_degree.sum()

    if total_edges == 0:
        return 0

    communities = np.array(communities)
    community_mat = communities[:, None] == communities[None, :]

    expected_matrix = np.outer(out_degree, in_degree) / total_edges
    modularity_matrix = (mat - expected_matrix) * community_mat
    return modularity_matrix.sum() / total_edges


@attrs.define(eq=True, frozen=True)
class SuggestedMerge:
    entity: str = attrs.field()
    target_community: str = attrs.field()
    gain: float = attrs.field(converter=float)

    def display(self) -> None:
        print(  # noqa: T201
            cyan(f"{self.entity}"),
            yellow("->"),
            cyan(f"{self.target_community}"),
            green(f"+{self.gain:.3f}"),
        )


@safe
def get_top_suggested_merges(adj_mat: AdjMat, top_n: int = 5) -> list[SuggestedMerge]:
    adj_mat = deepcopy(adj_mat)
    matrix: np.ndarray = adj_mat.mat
    communities: list[int] = adj_mat.communities.copy()
    comm_choices: set[int] = set(communities)

    starting_score = get_dwm(matrix, communities)

    suggested_merges = []

    for idx, comm in enumerate(communities):
        best_choice, best_score = None, starting_score

        for choice in comm_choices:
            if comm == choice:
                continue
            new_comms = communities.copy()
            new_comms[idx] = choice
            score = get_dwm(matrix, new_comms)

            if score > best_score:
                best_choice, best_score = choice, score

        if best_choice is not None:
            suggested_merges.append(
                SuggestedMerge(
                    adj_mat.node_map.get(idx),
                    adj_mat.comm_map.get(best_choice),
                    best_score - starting_score,
                )
            )

    return sorted(suggested_merges, key=lambda x: -x.gain)[:top_n]


def yellow(inp_str: str) -> str:
    return f"\033[33m{inp_str}\033[0m"


def cyan(inp_str: str) -> str:
    return f"\033[36m{inp_str}\033[0m"


def green(inp_str: str) -> str:
    return f"\033[32m{inp_str}\033[0m"
